import pydicom as dcm
import polars as pl

from datetime import date
from pathlib import Path
from loguru import logger


class EmptyReportError(FileNotFoundError):
    def __init__(self, message: str):
        super().__init__(message)


class ContentMissingError(ValueError):
    def __init__(self, message: str):
        super().__init__(message)


class AIRCReport:
    code_map = {
        "CHESTCT0203": "lung_parenchyma",
        "CHESTCT0304": "cardio",
        "CHESTCT0410": "aorta",
        "CHESTCT0502": "spine",
        "CHESTCT0611": "pulmonary_densities",
        "CHESTCT0999": "lesions",
    }
    lung_location_map = {
        "BothLungs": "both_lungs",
        "LeftUpperLobe": "left_upper_lobe",
        "LeftLowerLobe": "left_lower_lobe",
        "RightUpperLobe": "right_upper_lobe",
        "RightMiddleLobe": "right_middle_lobe",
        "RightLowerLobe": "right_lower_lobe",
    }
    finding_site_sequence = "363698007"
    tracking_code = "112039"

    def __init__(self, dicom_files: list[Path | str]):
        self.report_data = {}
        self.dicom_files = [Path(x) if not isinstance(x, Path) else x for x in dicom_files]
        self.series_uid = self.dicom_files[0].name.split('_')[0]

    def validate_dicoms(self):
        """Validate that all dicoms can be read properly and remove those that can't"""
        valid_dicoms = []
        for dicom in self.dicom_files:
            try:
                data = dcm.dcmread(dicom)
                valid_dicoms.append(data)
            except Exception as e:
                continue
        if not valid_dicoms:
            raise EmptyReportError("")
        self.dicom_data = valid_dicoms

    def extract_report(self) -> dict:
        """Extract the report data from the dicom files"""
        self.validate_dicoms()
        self.validate_identifiers()
        self.extract_measurements()
        logger.debug(f"{self.series_uid} AIRC Report extracted successfully")

    def validate_identifiers(self) -> None:
        """validate that the identifiers are present in the dicom data and are equal"""
        # Get reference values from first DICOM
        ref = self.dicom_data[0]
        ref_values = {
            "PatientID": ref.PatientID,
            "AccessionNumber": ref.AccessionNumber,
            "SeriesInstanceUID": ref.SeriesInstanceUID,
            "StudyInstanceUID": ref.StudyInstanceUID,
            "PatientSex": ref.PatientSex,
            "StudyDate": ref.StudyDate,
        }

        # Compare all other DICOMs against reference
        for data in self.dicom_data[1:]:
            for attr, ref_val in ref_values.items():
                curr_val = getattr(data, attr, None)
                if curr_val is None:
                    continue
                if curr_val != ref_val:
                    error_message = (
                        f"Mismatched {attr}: expected '{ref_val}', got '{curr_val}'"
                    )
                    logger.error(error_message)
                    raise ValueError(error_message)

        # Set the identifiers in the report data
        self.report_data["mrn"] = ref.PatientID
        self.report_data["accession"] = ref.AccessionNumber
        self.report_data["series_uid"] = ref.SeriesInstanceUID
        self.report_data["study_uid"] = ref.StudyInstanceUID
        self.report_data["sex"] = ref.PatientSex
        self.report_data["scan_date"] = date.fromisoformat(ref.StudyDate).strftime(
            "%Y-%m-%d"
        )
        self.report_data["extraction_date"] = date.today().isoformat()
        self.report_data["age"] = int(ref.PatientAge[:-1]) if ref.PatientAge != '' else None
        self.report_data["study_description"] = ref.StudyDescription

    def extract_measurements(self) -> None:
        for data in self.dicom_data:
            # try:
            #     measurement, measures = self._extract_measurement_from_dicom_data(data)
            #     if measures is not None:
            #         self.report_data[measurement] = measures
            # except Exception as e:
            #     logger.warning(
            #         f"Error extracting measurement from DICOM data: {e}"
            #     )
            #     continue
            measurement, measures = self._extract_measurement_from_dicom_data(data)
            if measures is not None:
                self.report_data[measurement] = measures
        self._merge_lung_data()
        self._create_main_dict()

    def _merge_lung_data(self):
        if (
            self.report_data.get("pulmonary_densities") is not None
            and self.report_data.get("lung_parenchyma") is not None
        ):
            # If we have both pulmonary densities and lung parenchyma, merge them
            combined_data = {}
            for location in self.lung_location_map.values():
                pulm_data = self.report_data["pulmonary_densities"].get(location, {})
                par_data = self.report_data["lung_parenchyma"].get(location, {})
                combined_data[location] = {**pulm_data, **par_data}
            self.report_data["lung"] = combined_data
            # Remove the old data
            del self.report_data["pulmonary_densities"]
            del self.report_data["lung_parenchyma"]
        # If we only have one of the two, rename it to lung_data
        else:
            if self.report_data.get("pulmonary_densities") is not None:
                self.report_data["lung"] = self.report_data["pulmonary_densities"]
                del self.report_data["pulmonary_densities"]
            elif self.report_data.get("lung_parenchyma") is not None:
                self.report_data["lung"] = self.report_data["lung_parenchyma"]
                del self.report_data["lung_parenchyma"]

    def _create_main_dict(self) -> None:
        """Create a dictionary of the main table data for the output database"""
        self.report_data["main"] = {
            "series_uid": self.report_data.get("series_uid"),
            "study_uid": self.report_data.get("study_uid"),
            "mrn": self.report_data.get("mrn"),
            "accession": self.report_data.get("accession"),
            "study_date": self.report_data.get("scan_date"),
            "extraction_date": self.report_data.get("extraction_date"),
            "study_descriptions": self.report_data.get("study_description"),
            "sex": self.report_data.get("sex"),
            "age": self.report_data.get("age"),
            "aorta": 1 if self.report_data.get("aorta") else 0,
            "spine": 1 if self.report_data.get("spine") else 0,
            "cardio": 1 if self.report_data.get("cardio") else 0,
            "lesions": 1 if self.report_data.get("lesions") else 0,
            "lung": 1 if self.report_data.get("lung") else 0,
        }

    def _extract_measurement_from_dicom_data(
        self, data: dcm.DataElement
    ) -> tuple[str, dict]:
        """Extract all appropriate AIRC measurements from one loaded dicom data
        :param data: dcm.DataElement
        :return: a tuple of the matched code and the measurement data
        """
        # data sequence
        data_content = self._check_for_content(data)
        measurement = self._match_code_to_airc_measurement(data_content)
        measure_content = self._get_measurement_content_sequence(data_content)
        # Get the measurement data
        match measurement:
            case "lung_parenchyma":
                measures = self._extract_lung_parenchyma_measurements(measure_content)
            case "cardio":
                measures = self._extract_coronary_calcium_measurements(measure_content)
            case "aorta":
                measures = self._extract_aortic_diameter_measurements(measure_content)
            case "spine":
                measures = self._extract_spine_measurements(measure_content)
            case "pulmonary_densities":
                measures = self._extract_pulmonary_density_measurements(measure_content)
            case "lesions":
                measures = self._extract_lung_lesion_measurements(measure_content)
        # Return the measurement data and the measurement name
        return measurement, measures

    def _check_for_content(self, data):
        if not hasattr(data, "ContentSequence"):
            logger.error(f"No ContentSequence found in {data.filename}")
            raise ContentMissingError("No ContentSequence found in DICOM data")

        content = data.ContentSequence
        self.current_filename = data.filename
        return content

    def _get_measurement_content_sequence(self, content):
        image_measure_code = "126010"
        measure_content = None
        for seq in content:
            if seq.ConceptNameCodeSequence[0].CodeValue == image_measure_code:
                # This is the image measure - we want to extract the data from this
                measure_content = seq
                break
        # If it's empty raise an error
        if not measure_content:
            logger.error(f"No image measure sequence found in {self.current_filename}")
            raise ContentMissingError("No image measure found in DICOM data")
        # If the sequence exists but doesn't have the content sequence, raise an error
        if not hasattr(measure_content, "ContentSequence"):
            logger.error(
                f"No measurement ContentSequence found in {self.current_filename}"
            )
            raise ContentMissingError("No ContentSequence found in DICOM data")
        return measure_content

    def _match_code_to_airc_measurement(self, content):
        id_content = content[0]
        code_map = AIRCReport.code_map
        if not hasattr(id_content, "ConceptCodeSequence"):
            logger.error(f"No AIRC Code found in {self.current_filename}")
            raise ContentMissingError("No AIRC Code found in DICOM data")
        # Match the code to the AIRC code map
        code = id_content.ConceptCodeSequence[0].CodeValue
        if code not in code_map:
            logger.error(
                f"Code {code} not found in AIRC code map for {self.current_filename}"
            )
            raise ContentMissingError("Code not found in AIRC code map")
        # This is one of the 6 AIRC measurements done - will be the key for the output dictionary
        measurement = code_map[code]
        return measurement

    def _extract_aortic_diameter_measurements(
        self, measure_content: dcm.DataElement
    ) -> dict:
        """Extract the aortic diameters from the dicom data
        :param content: the dicom data
        :return: a dictionary of the aortic diameters
        """
        # Get the measurements
        location_code_map = {
            "CHESTCT0408": "max_ascending",
            "CHESTCT0409": "max_descending",
            "C33557": "sinus_of_valsalva",
            "RID579": "sinotubular_junction",
            "CHESTCT0401": "mid_ascending",
            "CHESTCT0402": "proximal_arch",
            "CHESTCT0403": "mid_arch",
            "CHESTCT0404": "proximal_descending",
            "CHESTCT0405": "mid_descending",
            "CHESTCT0406": "diaphragm_level",
            "RID905": "celiac_artery_origin",
        }
        not_found_message = f"No aortic diameters found in {self.current_filename}"
        diameters = {}
        if not hasattr(measure_content, "ContentSequence"):
            logger.warning(not_found_message)
            return None
        aorta_measures = measure_content.ContentSequence
        for measure in aorta_measures:
            # Each measure is itself a sequence of data describing where the measure is taken and the value
            if not hasattr(measure, "ContentSequence"):
                continue
            measure_content = measure.ContentSequence
            diameter_sequence = "RID13432"
            site_location = None
            diameter = None
            # Loop through the sequences to pull out the location and the diameter
            for sequence in measure_content:
                seq_code = sequence.ConceptNameCodeSequence[0].CodeValue
                if seq_code == self.finding_site_sequence:
                    site_code = sequence.ConceptCodeSequence[0].CodeValue
                    # This is just the final code for PACS - not a meausurement
                    if site_code == "RID480":
                        continue
                    site_location = location_code_map.get(
                        site_code,
                        f"{sequence.ConceptCodeSequence[0].CodeValue}, {sequence.ConceptCodeSequence[0].CodeMeaning}",
                    )
                if seq_code == diameter_sequence:
                    # This is the measurement
                    diameter = int(sequence.MeasuredValueSequence[0].NumericValue)
            # If we have both the location and the diameter, add it to the dictionary
            if site_location is not None and diameter is not None:
                diameters[site_location] = diameter
        if not diameters:
            logger.warning(not_found_message)
            return None
        return diameters

    def _extract_lung_lesion_measurements(
        self, measure_content: dcm.DataElement
    ) -> dict:
        """Extract the lung lesion measurements from the dicom data
        :param measure_content: the dicom data
        :return: a dictionary of the lung lesion measurements
        """
        # Get the measurements
        lesion_data = {}
        not_found_message = f"No lung lesion measurements found in {self.current_filename}"
        if not hasattr(measure_content, "ContentSequence"):
            logger.warning(not_found_message)
            return None
        lesion_list = measure_content.ContentSequence
        # lesion_data['lesion_count'] = len(lesion_list)
        for idx, lesion in enumerate(lesion_list):
            lesion_id, lesion_measurements = self._extract_lung_lesion_measurement(
                lesion, idx
            )
            if lesion_id == 'No finding':
                # Skip this lesion as it is not a valid measurement
                continue
            if lesion_id is not None and lesion_measurements is not None:
                lesion_data[lesion_id] = lesion_measurements
        if not lesion_data:
            logger.warning(not_found_message)
            return None
        return lesion_data

    def _extract_lung_lesion_measurement(
        self, lesion: dcm.DataElement, idx: int
    ) -> tuple[str, dict]:
        if not hasattr(lesion, "ContentSequence"):
            logger.debug(
                f"No ContentSequence found in {self.current_filename} for lesion {idx}"
            )
            return None, None
        lesion_review_status_code = "CHESTCT0102"
        measurement_type_map = {
            "103339001": "max_2d_diameter_mm",
            "103340004": "min_2d_diameter_mm",
            "RID50155": "mean_2d_diameter_mm",
            "L0JK": "max_3d_diameter_mm",
            "RID28668": "volume_mm3",
        }
        lesion_measurements = {
            "location": None,
            "review_status": None,
            "max_2d_diameter_mm": None,
            "min_2d_diameter_mm": None,
            "mean_2d_diameter_mm": None,
            "max_3d_diameter_mm": None,
            "volume_mm3": None,
        }
        lobe_map = {
            "Upper lobe of left lung": "left_upper_lobe",
            "Lower lobe of left lung": "left_lower_lobe",
            "Upper lobe of right lung": "right_upper_lobe",
            "Middle lobe of right lung": "right_middle_lobe",
            "Lower lobe of right lung": "right_lower_lobe",
        }
        for seq in lesion.ContentSequence:
            descriptor = seq.ConceptNameCodeSequence[0]
            # Get the lesion ID
            if descriptor.CodeValue == self.tracking_code:
                lesion_id = seq.TextValue
            # Get the location
            if descriptor.CodeValue == self.finding_site_sequence:
                location = seq.ContentSequence[0].ConceptCodeSequence[0].CodeMeaning
                lesion_measurements["location"] = lobe_map.get(location, location)
            # Get the review status
            if descriptor.CodeValue == lesion_review_status_code:
                if seq.TextValue in (
                    "Measurement accepted",
                    "Measurement auto-confirmed",
                ):
                    review_status = "accepted"
                else:
                    review_status = seq.TextValue
                lesion_measurements["review_status"] = review_status
            if descriptor.CodeValue in measurement_type_map:
                measurement_type = measurement_type_map[descriptor.CodeValue]
                # Get the value
                if hasattr(seq, "MeasuredValueSequence"):
                    measurement_value = seq.MeasuredValueSequence[0].NumericValue
                    lesion_measurements[measurement_type] = float(measurement_value)
                else:
                    lesion_measurements[measurement_type] = None
        return lesion_id, lesion_measurements

    def _extract_lung_parenchyma_measurements(
        self, measure_content: dcm.DataElement
    ) -> dict:
        """Extract the lung parenchyma measurements from the dicom data
        :param measure_content: the dicom data
        :return: a dictionary of the lung parenchyma measurements
        """
        # Get the measurements
        not_found_message = f"No parenchyma measurements found in {self.current_filename}"
        if not hasattr(measure_content, "ContentSequence"):
            logger.warning(not_found_message)
            return None
        parenchyma_data = {}

        meausure_code = "CHESTCT0201"
        for location in measure_content.ContentSequence:
            location_content = location.ContentSequence
            for seq in location_content:
                descriptor = seq.ConceptNameCodeSequence[0]
                if descriptor.CodeValue == self.tracking_code:
                    location_id = self.lung_location_map.get(seq.TextValue)
                    if location_id is None:
                        continue
                if descriptor.CodeValue == meausure_code:
                    if not hasattr(seq, "MeasuredValueSequence"):
                        # If there is no measurement, skip this sequence
                        continue
                    # Get the measurement value
                    measurement_value = seq.MeasuredValueSequence[0].NumericValue
                    parenchyma_data[location_id] = {
                        "low_parenchyma_hu_percent": float(measurement_value)
                    }
        if not parenchyma_data:
            logger.warning(not_found_message)
            return None
        return parenchyma_data

    def _extract_coronary_calcium_measurements(
        self, measure_content: dcm.DataElement
    ) -> dict:
        """Extract the coronary calcium measurements from the dicom data
        :param measure_content: the dicom data
        :return: a dictionary of the coronary calcium measurements
        """
        not_found_message = f"No cardio measurements found in {self.current_filename}"
        calc_data = {}
        if not hasattr(measure_content, "ContentSequence"):
            logger.warning(not_found_message)
            return None
        for measure in measure_content.ContentSequence:
            measure_name = None
            measure_value = None
            if hasattr(measure, 'TextValue'):
                if measure.TextValue == 'No finding':
                    logger.warning(not_found_message)
                    return None
            for seq in measure.ContentSequence:
                if seq.ConceptNameCodeSequence[0].CodeValue == self.tracking_code:
                    # This is the location
                    if seq.TextValue == "Heart":
                        measure_name = "heart_volume_cm3"
                    elif seq.TextValue == "Calcium score":
                        measure_name = "coronary_calcification_volume_mm3"
                    # It appears that this "No finding" can appear in two locations?
                    elif seq.TextValue == 'No finding':
                        logger.warning(not_found_message)
                        return None
                if hasattr(seq, "MeasuredValueSequence"):
                    measure_value = seq.MeasuredValueSequence[0].NumericValue
            if measure_name is not None and measure_value is not None:
                calc_data[measure_name] = float(measure_value)
        if not calc_data:
            logger.warning(not_found_message)
            return None
        return calc_data

    def _extract_spine_measurements(self, measure_content: dcm.DataElement) -> dict:
        """Extract the spine measurements from the dicom data
        :param measure_content: the dicom data
        :return: a dictionary of the spine measurements
        """
        # Get the measurements
        spine_data = {}
        not_found_message = f"No spine measurements found in {self.current_filename}"
        if not hasattr(measure_content, "ContentSequence"):
            logger.warning(not_found_message)
            return None
        # Each sequence in this content is a vertebra's measurements
        for vertebra in measure_content.ContentSequence:
            vertebra_name, vertebra_measurements = self._extract_vertebra_measurement(
                vertebra
            )
            if vertebra_name is None or vertebra_measurements is None:
                continue
            # If we have a vertebra name and measurements, add them to the dictionary
            spine_data[vertebra_name] = vertebra_measurements
        if not spine_data:
            logger.warning(not_found_message)
            return None
        return spine_data

    def _extract_vertebra_measurement(
        self, vertebra: dcm.DataElement
    ) -> tuple[str, dict]:
        """Extract the vertebra measurements from the dicom data
        :param vertebra: the vertebra content sequence
        :return: a tuple of the vertebra name and the measurements
        """
        if not hasattr(vertebra, "ContentSequence"):
            logger.warning(
                f"No ContentSequence found in for a vertebra in {self.current_filename}"
            )
            return None, None
        # These are the internal codes used by the AIRC for the spine measurements
        measurement_seq_code = "121207"
        hounsfield_unit_code = "112031"
        direction_code = "106233006"
        status_code = "CHECTCT0001"

        vertebra_name = None
        vertebra_measurements = {}
        for seq in vertebra.ContentSequence:
            descriptor = seq.ConceptNameCodeSequence[0]
            if descriptor.CodeValue == self.tracking_code:
                vertebra_name = seq.TextValue
            if descriptor.CodeValue == measurement_seq_code:
                if not hasattr(seq, "MeasuredValueSequence"):
                    # If there is no measurement, skip this sequence
                    continue
                if not hasattr(seq, "ContentSequence"):
                    # If there is no content sequence, skip this sequence as we need to know the direction and status
                    continue
                # Get the measurement value
                measurement_value = seq.MeasuredValueSequence[0].NumericValue
                direction = None
                status = None
                for content in seq.ContentSequence:
                    if content.ConceptNameCodeSequence[0].CodeValue == direction_code:
                        direction = content.ConceptCodeSequence[0].CodeMeaning.lower()

                    if content.ConceptNameCodeSequence[0].CodeValue == status_code:
                        status = content.ConceptCodeSequence[0].CodeMeaning.lower()
                if direction is None or status is None:
                    # If we don't have a direction or status, skip this measurement
                    continue
                vertebra_measurements[direction] = {
                    "length_mm": float(measurement_value),
                    "status": status,
                }
            if descriptor.CodeValue == hounsfield_unit_code:
                if not hasattr(seq, "MeasuredValueSequence"):
                    continue
                measurement_value = seq.MeasuredValueSequence[0].NumericValue
                vertebra_measurements['mean_hu'] = float(measurement_value)
        if not vertebra_name or not vertebra_measurements:
            # logger.warning(
            #     f"No vertebra name or measurements found for a vertebra in {self.current_filename}"
            # )
            return None, None
        return vertebra_name, vertebra_measurements

    def _extract_pulmonary_density_measurements(
        self, measure_content: dcm.DataElement
    ) -> dict:
        """Extract the pulmonary density measurements from the dicom data
        :param measure_content: the dicom data
        :return: a dictionary of the pulmonary density measurements
        """
        # Get the measurements
        not_found_message = f"No pulmonary density measurements found in {self.current_filename}"
        if not hasattr(measure_content, "ContentSequence"):
            logger.warning(not_found_message)
            return None
        density_data = {}
        density_code_map = {
            "CHESTCT0601": "opacity_score",
            "CHESTCT0602": "volume_cm3",
            "CHESTCT0603": "opacity_volume_cm3",
            "CHESTCT0604": "opacity_percent",
            "CHESTCT0605": "high_opacity_volume_cm3",
            "CHESTCT0606": "high_opacity_percent",
            "CHESTCT0607": "mean_hu",
            "CHESTCT0608": "mean_hu_opacity",
        }
        for location in measure_content.ContentSequence:
            location_data = {}
            location_id = None
            location_content = location.ContentSequence
            for seq in location_content:
                descriptor = seq.ConceptNameCodeSequence[0]
                if descriptor.CodeValue == self.tracking_code:
                    location_id = self.lung_location_map.get(seq.TextValue)
                if descriptor.CodeValue in density_code_map:
                    meausure_name = density_code_map[descriptor.CodeValue]
                    if not hasattr(seq, "MeasuredValueSequence"):
                        # If there is no measurement, skip this sequence
                        continue
                    # Get the measurement value
                    measurement_value = seq.MeasuredValueSequence[0].NumericValue
                    if measurement_value == "n/a":
                        measurement_value = None
                    else:
                        measurement_value = float(measurement_value)
                    location_data[meausure_name] = measurement_value
            if location_id is not None and location_data:
                density_data[location_id] = location_data
        if not density_data:
            logger.warning(not_found_message)
            return None
        return density_data



import sqlite3
import polars as pl
from loguru import logger
from configparser import ConfigParser
from pathlib import Path

TABLE_COLUMNS = {
    "main": [
        "series_uid",
        "study_uid",
        "mrn",
        "accession",
        "study_date",
        "extraction_date",
        "study_description",
        "sex",
        "age",
        "aorta",
        "spine",
        "cardio",
        "lesions",
        "lung",
    ],
    "aorta": [
        "series_uid",
        "max_ascending",
        "max_descending",
        "sinus_of_valsalva",
        "sinotubular_junction",
        "mid_ascending",
        "proximal_arch",
        "mid_arch",
        "proximal_descending",
        "mid_descending",
        "diaphragm_level",
        "celiac_artery_origin",
    ],
    "spine": [
        "series_uid",
        "vertebra",
        "mean_hu",
        "direction",
        "length_mm",
        "status",
    ],
    "cardio": [
        "series_uid",
        "heart_volume_cm3",
        "coronary_calcification_volume_mm3",
    ],
    "lesions": [
        "series_uid",
        "lesion_id",
        "location",
        "review_status",
        "max_2d_diameter_mm",
        "min_2d_diameter_mm",
        "mean_2d_diameter_mm",
        "max_3d_diameter_mm",
        "volume_mm3",
    ],
    "lung": [
        "series_uid",
        "location",
        "opacity_score",
        "volume_cm3",
        "opacity_volume_cm3",
        "opacity_percent",
        "high_opacity_volume_cm3",
        "high_opacity_percent",
        "mean_hu",
        "mean_hu_opacity",
        "low_parenchyma_hu_percent",
    ],
}
# These are the columns that are not part of the primary key for each table
# We define these to allow .get() to work on the report data
DATA_COLUMNS = {
    "aorta": [
        "max_ascending",
        "max_descending",
        "sinus_of_valsalva",
        "sinotubular_junction",
        "mid_ascending",
        "proximal_arch",
        "mid_arch",
        "proximal_descending",
        "mid_descending",
        "diaphragm_level",
        "celiac_artery_origin",
    ],
    "spine": [
        "length_mm",
        "status",
    ],
    "cardio": [
        "heart_volume_cm3",
        "coronary_calcification_volume_mm3",
    ],
    "lesions": [
        "location",
        "review_status",
        "max_2d_diameter_mm",
        "min_2d_diameter_mm",
        "mean_2d_diameter_mm",
        "max_3d_diameter_mm",
        "volume_mm3",
    ],
    "lung": [
        "opacity_score",
        "volume_cm3",
        "opacity_volume_cm3",
        "opacity_percent",
        "high_opacity_volume_cm3",
        "high_opacity_percent",
        "mean_hu",
        "mean_hu_opacity",
        "low_parenchyma_hu_percent",
    ],
}


def create_new_data_db(data_db_path: Path | str) -> None:
    """
    Create a new output data database for AIRC data extraction with all required tables.
    :param data_db_path: Path to the new data database
    """
    main = """CREATE TABLE IF NOT EXISTS main (
        series_uid TEXT PRIMARY KEY,
        study_uid TEXT,
        mrn TEXT,
        accession TEXT,
        study_date TEXT,
        extraction_date TEXT,
        study_description TEXT,
        sex TEXT,
        age INTEGER,
        aorta INTEGER,
        spine INTEGER,
        cardio INTEGER,
        lesions INTEGER,
        lung INTEGER
    )"""
    aorta = """CREATE TABLE IF NOT EXISTS aorta (
        series_uid TEXT PRIMARY KEY,
        max_ascending INTEGER,
        max_descending INTEGER,
        sinus_of_valsalva INTEGER,
        sinotubular_junction INTEGER,
        mid_ascending INTEGER,
        proximal_arch INTEGER,
        mid_arch INTEGER,
        proximal_descending INTEGER,
        mid_descending INTEGER,
        diaphragm_level INTEGER,
        celiac_artery_origin INTEGER
    )"""
    spine = """CREATE TABLE IF NOT EXISTS spine (
        series_uid TEXT NOT NULL,
        vertebra TEXT NOT NULL,
        mean_hu REAL,
        direction TEXT,
        length_mm REAL,
        status TEXT,
        PRIMARY KEY (series_uid, vertebra, direction)
    )"""
    cardio = """CREATE TABLE IF NOT EXISTS cardio (
        series_uid TEXT PRIMARY KEY,
        heart_volume_cm3 REAL,
        coronary_calcification_volume_mm3 REAL
    )"""
    lesions = """CREATE TABLE IF NOT EXISTS lesions (
        series_uid TEXT NOT NULL,
        lesion_id TEXT NOT NULL,
        location TEXT,
        review_status TEXT,
        max_2d_diameter_mm REAL,
        min_2d_diameter_mm REAL,
        mean_2d_diameter_mm REAL,
        max_3d_diameter_mm REAL,
        volume_mm3 REAL,
        PRIMARY KEY (series_uid, lesion_id)
    )"""
    lung = """CREATE TABLE IF NOT EXISTS lung (
        series_uid TEXT NOT NULL,
        location TEXT NOT NULL,
        opacity_score REAL,
        volume_cm3 REAL,
        opacity_volume_cm3 REAL,
        opacity_percent REAL,
        high_opacity_volume_cm3 REAL,
        high_opacity_percent REAL,
        mean_hu REAL,
        mean_hu_opacity REAL,
        low_parenchyma_hu_percent REAL,
        PRIMARY KEY (series_uid, location)
    )"""
    with sqlite3.connect(data_db_path) as conn:
        cursor = conn.cursor()
        for table in [main, aorta, spine, cardio, lesions, lung]:
            try:
                cursor.execute(table)
            except sqlite3.Error as e:
                logger.error(f"Error creating table: {e}")
                print(table)
        conn.commit()
        logger.success(
            f"Created new data database at {data_db_path} with required tables."
        )


def query_unextracted_data(config: ConfigParser) -> list[list]:
    """
    Query the DicomConquest database for unextracted dicom files.
    :param config: Configuration object
    :return: List of lists. Each inner list contains the dicom file paths for a single study.
    """
    dicom_db = config.get("GENERAL", "dicom_db")
    data_db = config.get("GENERAL", "data_db")
    conn = sqlite3.connect(dicom_db)
    with conn:
        cursor = conn.cursor()
        cursor.execute(f"ATTACH DATABASE '{data_db}' AS data_db")
        query = """SELECT main.DICOMImages.SeriesInst as series_uid, main.DICOMImages.ObjectFile as filepath
                FROM main.DICOMImages
                INNER JOIN main.DICOMSeries
                ON main.DICOMImages.SeriesInst = main.DICOMSeries.SeriesInst
                LEFT JOIN data_db.main ON main.DICOMImages.SeriesInst = data_db.main.series_uid
                WHERE data_db.main.series_uid IS NULL
                AND main.DICOMSeries.Modality = 'SR'
                """

        cursor.execute(query)
        unextracted = (
            pl.read_database(query, conn)
            .group_by("series_uid")
            .agg(pl.col("filepath").unique().alias("filepaths"))["filepaths"]
            .to_list()
        )
        cursor.execute("DETACH DATABASE data_db")
    conn.close()
    return unextracted


def insert_data_to_db(report, config: ConfigParser) -> None:
    """
    Insert AIRC Report into the database.
    :param report: AIRC Report Object
    :param config: Configuration object
    """
    airc_data = report.report_data
    data_db = config.get("GENERAL", "data_db")
    conn = sqlite3.connect(data_db)
    with conn:
        for table in TABLE_COLUMNS:
            if table not in airc_data:
                logger.debug(
                    f"{table.title()} data not found in {report.series_uid}. Skipping database insert."
                )
                continue
            insert_statement = get_insert_statement(table)
            formatted_data = format_table_input(airc_data, table)
            try:
                conn.executemany(insert_statement, formatted_data)
                conn.commit()
            except sqlite3.Error as e:
                logger.error(f"Error inserting {report.series_uid} into {table}: {e}")
    conn.close()
    logger.debug(f"{report.series_uid} inserted into database.")


def format_table_input(report_data: dict, table_name: str) -> tuple:
    """
    Format the input data for a given table.
    :param report_data: Dictionary containing the report data
    :param table_name: Name of the table to format the data for
    :return: Tuple of formatted data
    """
    data_cols = DATA_COLUMNS.get(table_name)
    match table_name:
        case "main":
            formatted = [tuple(report_data.get(table_name).values())]
        case "lesions":
            formatted = []
            for lesion, data in report_data.get("lesions").items():
                row = (
                    report_data.get("series_uid"),
                    lesion,
                    *[data.get(col) for col in data_cols],
                )
                formatted.append(row)
        case "spine":
            formatted = []
            for vertebra, measurements in report_data.get("spine").items():
                mean_hu = measurements.pop('mean_hu', None)
                for direction, data in measurements.items():
                    row = (
                        report_data.get("series_uid"),
                        vertebra,
                        mean_hu,
                        direction,
                        *[data.get(col) for col in data_cols],
                    )
                    formatted.append(row)
        case "lung":
            formatted = []
            for location, data in report_data.get("lung").items():
                row = (
                    report_data.get("series_uid"),
                    location,
                    *[data.get(col) for col in data_cols],
                )
                formatted.append(row)
        case _:
            formatted = [
                (
                    report_data.get("series_uid"),
                    *[report_data[table_name].get(col) for col in data_cols],
                )
            ]
    return formatted


def get_insert_statement(table_name: str) -> str:
    """
    Get the insert statement for a given table name.
    :param table_name: Name of the table
    :return: Insert statement
    """
    columns = ", ".join(TABLE_COLUMNS[table_name])
    placeholders = ", ".join(["?"] * len(TABLE_COLUMNS[table_name]))
    return f"REPLACE INTO {table_name} ({columns}) VALUES({placeholders})"



import os
import sys
import argparse
import configparser
import sqlite3

from datetime import datetime
from loguru import logger
from pathlib import Path, PureWindowsPath
from airc_extract.db_ops import (
    create_new_data_db,
    query_unextracted_data,
    insert_data_to_db,
)
from airc_extract.airc_report import AIRCReport, EmptyReportError


def main() -> None:
    parser = argparse.ArgumentParser(description="AIRC Data Extractor")
    parser.add_argument(
        "--config",
        "-c",
        type=str,
        default=Path(__file__).resolve().parent / "config.ini",
        help="Path to the configuration file. If you ran create_airc_config, this is will point to the default config file.",
    )
    args = parser.parse_args()

    config = _load_config(args.config)
    _setup_logging(config)
    _test_connections(config)
    airc_data_extractor(config)


def airc_data_extractor(config: configparser.ConfigParser) -> None:
    """
    Main function to extract AIRC data.
    :param config: Configuration object
    """
    logger.info("Starting AIRC data extraction...")
    data_dir = Path(config.get("GENERAL", "dicom_data_dir"))
    unextracted_studies = query_unextracted_data(config)
    total_studies = len(unextracted_studies)
    logger.info(f"Found {total_studies} unextracted studies in the DICOM database.")
    successes = 0
    for i, study in enumerate(unextracted_studies, start=1):
        try:
            study = [data_dir / Path(file) for file in study]
            report = AIRCReport(study)
            report.extract_report()
            insert_data_to_db(report, config)
        except EmptyReportError as e:
            logger.error(
                f"{i}/{total_studies} - {report.series_uid} has no valid dicom files. Skipping extraction. {e}"
            )
            continue
        except Exception as e:
            logger.critical(f"{i}/{total_studies} - {report.series_uid} failed: {e}")
            continue
        logger.success(
            f"{i}/{total_studies} - {report.series_uid} extracted and inserted into database."
        )
        successes += 1
    logger.success(f'Extraction completed. Successfully inserted {successes} / {total_studies} into output database.')


def _setup_logging(config: configparser.ConfigParser) -> None:
    """
    Set up logging for the AIRC data extractor.
    """
    log_dir = Path(config.get("GENERAL", "log_dir"))
    log_level = config.get("GENERAL", "log_level")
    log_level_file = config.get("GENERAL", "log_level_file")
    today = datetime.today().strftime("%Y_%m_%d")
    # Set up terminal logging
    logger.remove()
    logger.add(
        sys.stderr,
        level=log_level,
        enqueue=True,
        format="<green>{time:YYYY-MM-DD at HH:mm:ss}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>",
    )
    # Set up file logging
    logger.add(
        Path(log_dir) / f"airc_data_extractor_{today}.log",
        level=log_level_file,
        rotation="1 GB",
        enqueue=True,
        format="{time:YYYY-MM-DD at HH:mm:ss} | {level: <8} | {name}:{function}:{line} - {message}",
    )


def _load_config(config_path: str) -> configparser.ConfigParser:
    # Set up the configuration parser
    config_path = Path(config_path).resolve()
    if not config_path.exists():
        raise FileNotFoundError(
            f"Configuration file {config_path} not found. Please run create_airc_config to create it."
        )
    # Load the configuration file
    config = configparser.ConfigParser()
    config.read(config_path)
    return config


def _test_connections(config) -> None:
    """
    Test the connections to the DICOM and data databases.
    """
    # Placeholder for actual connection testing logic
    logger.debug("Testing connections to DICOM and data databases...")
    dicom_db = config.get("GENERAL", "dicom_db")
    data_db = config.get("GENERAL", "data_db")
    dicom_db = Path(dicom_db)
    data_db = Path(data_db)
    # If the database paths ar
    if not dicom_db.exists():
        raise FileNotFoundError(f"DICOM database {dicom_db} not found.")
    if not data_db.exists():
        raise FileNotFoundError(f"Data database {data_db} not found.")

    # Test the connections
    with sqlite3.connect(dicom_db) as conn, sqlite3.connect(data_db) as conn2:
        try:
            cursor = conn.cursor()
            logger.info("Connected to DICOM database.")
        except Exception as e:
            raise sqlite3.Error(f"Error connecting to dicom database {dicom_db}: {e}")
        # Check data database connection
        try:
            cursor = conn2.cursor()
            logger.info("Connected to data database.")
        except Exception as e:
            raise sqlite3.Error(f"Error connecting to data database {data_db}: {e}")


def create_airc_config() -> None:
    """
    Create a configuration file for the AIRC data extractor.
    """
    parser = argparse.ArgumentParser(description="Create AIRC configuration file")
    parser.add_argument(
        "--dicom-db",
        "-d",
        type=str,
        required=True,
        help="Path to the DicomConquest database",
    )
    parser.add_argument(
        "--dicom-data-dir",
        "-dd",
        type=str,
        required=True,
        help="Path to the DicomConquest data directory",
    )
    parser.add_argument(
        "--data-db", "-o", type=str, required=True, help="Path to the data database"
    )
    parser.add_argument(
        "--log-level-term",
        "-l",
        type=str,
        default="INFO",
        help="Terminal logging level: DEBUG, INFO, WARNING, ERROR, CRITICAL",
    )
    parser.add_argument(
        "--log-level-file",
        "-f",
        type=str,
        default="DEBUG",
        help="File logging level: DEBUG, INFO, WARNING, ERROR, CRITICAL",
    )
    parser.add_argument(
        "--log-dir", type=str, default=".", help="Directory for log files"
    )
    args = parser.parse_args()
    lib_path = Path(__file__).resolve().parent
    config_path = lib_path / "config.ini"
    config = configparser.ConfigParser()
    dicom_db = Path(args.dicom_db).resolve().absolute()
    dicom_data_dir = Path(args.dicom_data_dir).resolve().absolute()
    data_db = Path(args.data_db).resolve().absolute()
    log_dir = Path(args.log_dir).resolve().absolute()
    config["GENERAL"] = {
        "dicom_db": str(dicom_db),
        "dicom_data_dir": str(dicom_data_dir),
        "data_db": str(data_db),
        "log_dir": log_dir,
        "log_level": args.log_level_term,
        "log_level_file": args.log_level_file,
    }
    with open(config_path, "w") as configfile:
        config.write(configfile)

    if not data_db.exists():
        create_new_data_db(data_db)
    else:
        logger.info(f"Data database {data_db} already exists. Skipping creation.")
    if not dicom_db.exists():
        raise FileNotFoundError(f"DICOM database {dicom_db} not found.")
    if not log_dir.exists():
        log_dir.mkdir(parents=True, exist_ok=False)
    _test_connections(config)
    logger.success(
        f"""Created configuration file at {config_path} with the following settings:
        DICOM database: {dicom_db}
        DICOM data directory: {dicom_data_dir}
        Data database: {data_db}
        Log directory: {log_dir}
        Terminal Log level: {args.log_level_term}
        File Log level: {args.log_level_file}"""
    )
